#include "opendrive_link.h"
#include "opendrive_road_node.h"

COpendriveLink::COpendriveLink()
{
  this->prev=NULL;
  this->next=NULL;
  this->spline=NULL;
  this->mark=OD_MARK_NONE;
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
}

COpendriveLink::COpendriveLink(const COpendriveLink &object)
{
  this->prev=object.prev;
  this->next=object.next;
  this->spline=new CG2Spline(*object.spline);
  this->mark=object.mark;
  this->resolution=object.resolution;
  this->scale_factor=object.scale_factor;
}

void COpendriveLink::set_prev(COpendriveRoadNode *node)
{
  this->prev=node;
}

void COpendriveLink::set_next(COpendriveRoadNode *node)
{
  this->next=node;
}

void COpendriveLink::set_road_mark(opendrive_mark_t mark)
{
  this->mark=mark;
}

void COpendriveLink::set_resolution(double res)
{
  this->resolution=res;
}

void COpendriveLink::set_scale_factor(double scale)
{
  this->scale_factor=scale;
}

void COpendriveLink::generate(double start_curvature,double end_curvature,double length,bool lane)
{
  TOpendriveWorldPoint node_start,node_end;
  TPoint start,end;

  node_start=this->prev->get_start_pose();
  start.x=node_start.x;
  start.y=node_start.y;
  start.heading=node_start.heading;
  start.curvature=start_curvature;
  node_end=this->next->get_start_pose();
  end.x=node_end.x;
  end.y=node_end.y;
  end.heading=node_end.heading;
  end.curvature=end_curvature;
//  std::cout << start.x << "," << start.y << "," << start.heading << "," << start_curvature << "," << end.x << "," << end.y << "," << end.heading << "," << end_curvature << std::endl;
  this->spline=new CG2Spline(start,end);
  this->spline->generate(this->resolution,length);
}

void COpendriveLink::update_references(std::map<COpendriveRoadNode *,COpendriveRoadNode *> refs)
{
  this->prev=refs[this->prev];
  this->next=refs[this->next];
}

const COpendriveRoadNode &COpendriveLink::get_prev(void) const
{
  return *this->prev;
}

const COpendriveRoadNode &COpendriveLink::get_next(void) const
{
  return *this->next;
}

opendrive_mark_t COpendriveLink::get_road_mark(void) const
{
  return this->mark;
}

double COpendriveLink::get_resolution(void) const
{
  return this->resolution;
}

double COpendriveLink::get_scale_factor(void) const
{
  return this->scale_factor;
}

double COpendriveLink::find_closest_world_point(TOpendriveWorldPoint &world,TPoint &point)
{

}

double COpendriveLink::find_closest_local_point(TOpendriveLocalPoint &local,TPoint &point)
{

}

double COpendriveLink::get_track_point_world(TOpendriveTrackPoint &track,TOpendriveWorldPoint &world)
{

}

double COpendriveLink::get_track_point_local(TOpendriveTrackPoint &track,TOpendriveLocalPoint &local)
{

}

void COpendriveLink::get_trajectory(std::vector<double> &x, std::vector<double> &y,double start_length, double end_length) const
{
  std::vector<double> curvature,heading;
  CG2Spline *partial_spline=NULL;
  
  if(start_length!=0.0 || end_length!=-1.0)// get partial spline
  {
    partial_spline=new CG2Spline;
    this->spline->get_part(partial_spline,start_length,end_length);
    partial_spline->evaluate_all(x,y,curvature,heading);
    delete partial_spline;
  }
  else
    this->spline->evaluate_all(x,y,curvature,heading);
}

void COpendriveLink::get_trajectory(std::vector<double> &x, std::vector<double> &y, std::vector<double> &yaw,double start_length, double end_length) const
{
  std::vector<double> curvature;
  CG2Spline *partial_spline=NULL;

  if(start_length!=0.0 || end_length!=-1.0)// get partial spline
  {
    partial_spline=new CG2Spline;
    this->spline->get_part(partial_spline,start_length,end_length);
    partial_spline->evaluate_all(x,y,curvature,yaw);
    delete partial_spline;
  }
  else
    this->spline->evaluate_all(x,y,curvature,yaw);
}

double COpendriveLink::get_length(void) const
{
  if(this->spline!=NULL)
    return this->spline->get_length();
  else
    return 0.0;
}

std::ostream& operator<<(std::ostream& out, COpendriveLink &link)
{
  out << "        Previous node: " << link.get_prev().get_index() << " of road " << link.get_prev().get_parent_segment().get_name() << std::endl;
  out << "        Next node: " << link.get_next().get_index() << " of road " << link.get_next().get_parent_segment().get_name() << std::endl;
  out << "        Road mark: ";
  switch(link.get_road_mark())
  {
    case OD_MARK_NONE:
      out << "no mark" << std::endl;
      break;
    case OD_MARK_SOLID:
      out << "solid" << std::endl;
      break;
    case OD_MARK_BROKEN:
      out << "broken" << std::endl;
      break;
    case OD_MARK_SOLID_SOLID:
      out << "solid solid" << std::endl;
      break;
    case OD_MARK_SOLID_BROKEN:
      out << "solid broken" << std::endl;
      break;
    case OD_MARK_BROKEN_SOLID:
      out << "broken solid" << std::endl;
      break;
    case OD_MARK_BROKEN_BROKEN:
      out << "broken broken" << std::endl;
      break;
  }

  return out;
}

COpendriveLink::~COpendriveLink()
{
  this->prev=NULL;
  this->next=NULL;
  if(this->spline!=NULL)
  {
    delete this->spline;
    this->spline=NULL;
  }
  this->mark=OD_MARK_NONE;
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
}

