#include "opendrive_road.h"
#include <sys/types.h>
#include <sys/stat.h>
#include "exceptions.h"

COpendriveRoad::COpendriveRoad()
{
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
  this->min_road_length=DEFAULT_MIN_ROAD_LENGTH;
  this->segments.clear();
  this->nodes.clear();
}

COpendriveRoad::COpendriveRoad(const COpendriveRoad& object)
{
  COpendriveRoadSegment *segment;
  COpendriveRoadNode *node;
  std::map<COpendriveRoadSegment *,COpendriveRoadSegment *> new_segment_ref;
  std::map<COpendriveRoadNode *,COpendriveRoadNode *> new_node_ref;

  this->resolution=object.resolution;
  this->scale_factor=object.scale_factor;
  this->min_road_length=object.min_road_length;
  this->segments.clear();
  for(unsigned int i=0;i<object.nodes.size();i++)
  {
    node=new COpendriveRoadNode(*object.nodes[i]);
    this->nodes.push_back(node);
    new_node_ref[object.nodes[i]]=node;
  }
  // update references
  for(unsigned int i=0;i<this->nodes.size();i++)
    this->nodes[i]->update_references(new_node_ref);

  for(unsigned int i=0;i<object.segments.size();i++)
  {
    segment=new COpendriveRoadSegment(*object.segments[i],new_node_ref,this);
    this->segments.push_back(segment);
    new_segment_ref[object.segments[i]]=segment;
  }
  this->nodes.clear();
  // update references
  for(unsigned int i=0;i<this->segments.size();i++)
    this->segments[i]->update_references(new_segment_ref);
  for(unsigned int i=0;i<this->nodes.size();i++)
    this->nodes[i]->update_references(new_segment_ref);
}

COpendriveRoadSegment &COpendriveRoad::operator[](std::string &key)
{
  for(unsigned int i=0;i<this->segments.size();i++)
  {
    if(this->segments[i]->get_id()==(unsigned int)std::stoi(key))
      return *this->segments[i];
  }

  throw CException(_HERE_,"No road segment with the given ID");
}

void COpendriveRoad::link_segments(OpenDRIVE &open_drive)
{
  std::string predecessor_id,successor_id;

  for(OpenDRIVE::road_iterator road_it(open_drive.road().begin());road_it!=open_drive.road().end();++road_it)
  {
    // get current segment
    COpendriveRoadSegment &road=(*this)[road_it->id().get()];
    // get predecessor and successor
    if(road_it->lane_link().present())
    {
      if(road_it->lane_link().get().predecessor().present())// predecessor present
      {
        if(road_it->lane_link().get().predecessor().get().elementType().get()=="road")// previous segment is a road
          predecessor_id=road_it->lane_link().get().predecessor().get().elementId().get();
      }
      if(road_it->lane_link().get().successor().present())// successor present
      {
        if(road_it->lane_link().get().successor().get().elementType().get()=="road")
          successor_id=road_it->lane_link().get().successor().get().elementId().get();
      }
    }
    if(std::stoi(road_it->junction().get())==-1)// non junction road segments
    {
      if(!predecessor_id.empty())
      {
        COpendriveRoadSegment &prev_road=(*this)[predecessor_id];
        prev_road.link_segment(road);
        predecessor_id.clear();
      }
      if(!successor_id.empty())
      {
        COpendriveRoadSegment &next_road=(*this)[successor_id];
        road.link_segment(next_road);
        successor_id.clear();
      }
    }
    else// junction segment
    {
      for(OpenDRIVE::junction_iterator junction_it(open_drive.junction().begin());junction_it!=open_drive.junction().end();++junction_it)
      {
        for(junction::connection_iterator connection_it(junction_it->connection().begin()); connection_it!=junction_it->connection().end();++connection_it)
        {
          std::string incoming_road_id;
          std::string connecting_road_id;
          if(connection_it->incomingRoad().present())
            incoming_road_id=connection_it->incomingRoad().get();
          else
            throw CException(_HERE_,"Connectivity information missing");
          if(connection_it->connectingRoad().present())
            connecting_road_id=connection_it->connectingRoad().get();
          else
            throw CException(_HERE_,"Connectivity information missing");
          if(predecessor_id.compare(incoming_road_id)==0 && successor_id.compare(connecting_road_id)==0)// this is the connection
          {
            COpendriveRoadSegment &prev_road=(*this)[predecessor_id];
            COpendriveRoadSegment &next_road=(*this)[successor_id];
            for(connection::laneLink_iterator lane_link_it(connection_it->laneLink().begin()); lane_link_it!=connection_it->laneLink().end();++lane_link_it)
            {
              int from_lane_id;
              int to_lane_id;
              if(lane_link_it->from().present())
                from_lane_id=lane_link_it->from().get();
              else
                throw CException(_HERE_,"Connectivity information missing");
              if(lane_link_it->to().present())
                to_lane_id=lane_link_it->to().get();
              else
                throw CException(_HERE_,"Connectivity information missing");
              prev_road.link_segment(road,from_lane_id,-1);
              road.link_segment(next_road,-1,to_lane_id);
            }
          }
        }
      }
    }
  }
}

unsigned int COpendriveRoad::add_node(COpendriveRoadNode *node)
{
  for(unsigned int i=0;i<this->nodes.size();i++)
  {
    if(this->nodes[i]==node)
      throw CException(_HERE_,"Node already present");
  }
  this->nodes.push_back(node);

  return this->nodes.size()-1;
}

bool COpendriveRoad::node_exists_at(const TOpendriveWorldPoint &pose)
{
  TOpendriveWorldPoint node_pose;

  for(unsigned int i=0;i<nodes.size();i++)
  {
    node_pose=this->nodes[i]->get_start_pose();
    if(std::abs(pose.x-node_pose.x)<0.01 && std::abs(pose.y-node_pose.y)<0.01)
      return true;
  }

  return false;
}

COpendriveRoadNode* COpendriveRoad::get_node_at(const TOpendriveWorldPoint &pose) const
{
  TOpendriveWorldPoint node_pose;

  for(unsigned int i=0;i<nodes.size();i++)
  {
    node_pose=this->nodes[i]->get_start_pose();
    if(std::abs(pose.x-node_pose.x)<0.01 && std::abs(pose.y-node_pose.y)<0.01)
      return this->nodes[i];
  }

  return NULL;
}

void COpendriveRoad::load(const std::string &filename)
{
  struct stat buffer;

  if(stat(filename.c_str(),&buffer)==0)
  {
    // delete any previous data
    for(unsigned int i=0;i<this->segments.size();i++)
      delete this->segments[i];
    this->segments.clear();
    // try to open the specified file
    try{
      std::unique_ptr<OpenDRIVE> open_drive(OpenDRIVE_(filename.c_str(), xml_schema::flags::dont_validate));
      for (OpenDRIVE::road_iterator road_it(open_drive->road().begin()); road_it != open_drive->road().end(); ++road_it)
      {
        try{
          COpendriveRoadSegment *segment=new COpendriveRoadSegment();
          segment->set_resolution(this->resolution);
          segment->set_scale_factor(this->scale_factor);
          segment->set_min_road_length(this->min_road_length);
          segment->set_parent_road(this);
          segment->load(*road_it);
          this->segments.push_back(segment);
        }catch(CException &e){
          std::cout << e.what() << std::endl;
        }
      }
      // link segments
      this->link_segments(*open_drive);
      // process junctions
    }catch (const xml_schema::exception& e){
      std::ostringstream os;
      os << e;
      /* handle exceptions */
      throw CException(_HERE_,os.str());
    }
  }
  else
    throw CException(_HERE_,"The .xodr file does not exist");
}

void COpendriveRoad::set_resolution(double res)
{
  this->resolution=res;

  // change the resolution of all current nodes
  for(unsigned int i=0;i<this->segments.size();i++)
    this->segments[i]->set_resolution(res);
}

double COpendriveRoad::get_resolution(void) const
{
  return this->resolution;
}

void COpendriveRoad::set_scale_factor(double scale)
{
  this->scale_factor=scale;

  // change the resolution of all current nodes
  for(unsigned int i=0;i<this->segments.size();i++)
    this->segments[i]->set_scale_factor(scale);
}

double COpendriveRoad::get_scale_factor(void) const
{
  return this->scale_factor;
}

void COpendriveRoad::set_min_road_length(double length)
{
  this->min_road_length=length;

  for(unsigned int i=0;i<this->segments.size();i++)
    this->segments[i]->set_min_road_length(length);
}

double COpendriveRoad::get_min_road_length(void) const
{
  return this->min_road_length;
}

unsigned int COpendriveRoad::get_num_segments(void) const
{
  return this->segments.size();
}

const COpendriveRoadSegment& COpendriveRoad::get_segment(unsigned int index) const
{
  if(index>=0 && index<this->segments.size())
    return *this->segments[index];
  else
    throw CException(_HERE_,"Invalid segment index");
}

unsigned int COpendriveRoad::get_num_nodes(void) const
{
  return this->nodes.size();
}

const COpendriveRoadNode& COpendriveRoad::get_node(unsigned int index) const
{
  if(index>=0 && index<this->nodes.size())
    return *this->nodes[index];
  else
    throw CException(_HERE_,"Invalid node index");
}

const COpendriveRoadSegment& COpendriveRoad::operator[](std::size_t index)
{
  if(index>=0 && index<this->segments.size())
    return *this->segments[index];
  else
    throw CException(_HERE_,"Invalid segment index");
}

void COpendriveRoad::operator=(const COpendriveRoad& object)
{
  COpendriveRoadSegment *segment;
  COpendriveRoadNode *node;
  std::map<COpendriveRoadSegment *,COpendriveRoadSegment *> new_references;

  this->resolution=object.resolution;
  this->scale_factor=object.scale_factor;
  this->min_road_length=object.min_road_length;
  for(unsigned int i=0;i<this->segments.size();i++)
  {
    delete this->segments[i];
    this->segments[i]=NULL;
  }
  this->segments.clear();
  for(unsigned int i=0;i<object.segments.size();i++)
  {
    segment=new COpendriveRoadSegment(*object.segments[i]);
    this->segments.push_back(segment);
    new_references[object.segments[i]]=segment;
  }
  for(unsigned int i=0;i<this->nodes.size();i++)
  {
    delete this->nodes[i];
    this->nodes[i]=NULL;
  }
  this->nodes.clear();
  for(unsigned int i=0;i<object.nodes.size();i++)
  {
    node=new COpendriveRoadNode(*object.nodes[i]);
    this->nodes.push_back(node);
  }
  for(unsigned int i=0;i<this->segments.size();i++)
    this->segments[i]->update_references(new_references);
}

std::ostream& operator<<(std::ostream& out, COpendriveRoad &road)
{
  out << "Resolution: " << road.resolution << std::endl;
  out << "Scale factor: " << road.scale_factor << std::endl;
  out << "Minimum road lenght: " << road.min_road_length << std::endl;
  out << "Total number of nodes: " << road.nodes.size() << std::endl;
  for(unsigned int i=0;i<road.segments.size();i++)
    std::cout << *road.segments[i];

  return out;
}

COpendriveRoad::~COpendriveRoad()
{
  for(unsigned int i=0;i<this->segments.size();i++)
  {
    delete this->segments[i];
    this->segments[i]=NULL;
  }
  for(unsigned int i=0;i<this->nodes.size();i++)
  {
    delete this->nodes[i];
    this->nodes[i]=NULL;
  }
  this->segments.clear();
  this->resolution=DEFAULT_RESOLUTION;
  this->scale_factor=DEFAULT_SCALE_FACTOR;
  this->min_road_length=DEFAULT_MIN_ROAD_LENGTH;
}

